
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <getopt.h>
#include "conf.h"

#define AUTHOR		"Stephan Baerwolf, stephan@matrixstorm.com"
#define VERSION		"1.0.1"
#define DESCRIPTION	"gives non-root useres the power to change their group-cfs usershare"
#define NAME		"sharetool"
#define	SOFTWARE	NAME ", v" VERSION " by " AUTHOR


// internal used
//#define DEBUG_SHARETOOL
#define	SHAREPATH	"/sys/kernel/uids" 




static struct conf_schedtool globalconf;




#define conf_schedtool_getsharefilename(a) ({									\
	char buffer[CONF_SCHEDTOOL_SHARENAMEMAXLEN + CONF_SCHEDTOOL_SHARENAMEMAXLEN + 1] = "";			\
	sprintf(buffer, SHAREPATH "/" "%s" "/cpu_share" , (a)->sharename);						\
	buffer;													\
})

static void usage(FILE *device) {
	fprintf(device, SOFTWARE "\n");
	fprintf(device, NAME " " DESCRIPTION "\n\n");

	fprintf(device, NAME " [<options>] [[--value] <new sharevalue>] \n\n"); 
	fprintf(device, "	--onlyshow, -s		shows the selected share only \n");
	fprintf(device, "	--sharename, -N	<name>	selects ths share \"name\" \n");
	fprintf(device, "	--value	<sharevalue>	defines the new share-value \n");

	fprintf(device, "\n\n");

	fflush(device);
	return;
} 

static int isAllowedEx(struct conf_schedtool *conf) {
	char buffer[256] = "";
	uid_t uid = getuid();	// we do not use conf->userid BECAUSE: safety first !

	if (uid == 0) return 1; // root is always allowed

	// user is allowed to (only!) REDUCE its own share ...
	// ...if this process has effective root-priviledges
	if (uid == conf->userid) {
		sprintf(buffer,"%i",uid);
		if (strncmp(conf->sharename, buffer, CONF_SCHEDTOOL_SHARENAMEMAXLEN) == 0)
		 if (conf->shareval <= conf->current_shareval)
		  if (conf->shareval >= CONF_SCHEDTOOL_MINUSERSHAREVAL)
		   if (geteuid() == 0) return 1; //at least we net effective root-user priviledges (userbit)
	} //ENDE: normaluser safety first !

	return 0;
} 

static int currentShareValueEx(struct conf_schedtool *conf) {
	char	sharevalstr[256] = "";
	FILE*	f	= fopen(conf_schedtool_getsharefilename(conf), "r");
	int	result	= -1;

	if (f != NULL) {
		result=atoi(fgets(sharevalstr, sizeof(sharevalstr), f));
		fclose(f);
	}

	if (result < CONF_SCHEDTOOL_MINSHAREVAL)	return -1;
	else						return result;
}

static int  updateShareValueEx(struct conf_schedtool *conf) {
	char	sharevalstr[256] = "";
	FILE*	f	= NULL;
	int	result	= -1;

	if (conf->shareval < CONF_SCHEDTOOL_MINSHAREVAL) return -2;

	if (!((conf->current_shareval=currentShareValueEx(conf)) < 0)) 
	 if (conf->current_shareval != conf->shareval) {
		sprintf(sharevalstr, "%i", conf->shareval);
		f=fopen(conf_schedtool_getsharefilename(conf), "w");	
		if (f != NULL) {
			result=fputs(sharevalstr, f); // write the new value !
			fclose(f);

			conf->current_shareval=currentShareValueEx(conf);
		}
	 } else result=0;

	if (result < 0)	return -1;
	else		return 0;
}

static int isAllowed(void) {
	return isAllowedEx(&globalconf);
}

static int currentShareValue(void) {
	return currentShareValueEx(&globalconf);
}

static int  updateShareValue(void) {
	return updateShareValueEx(&globalconf);
}

static int parseconf(struct conf_schedtool *conf, int argc, char** argv) {
	static struct option long_opts[] = {
		{"help", 0, 0, 'h'},
		{"onlyshow", 0, 0, 's'},
		{"sharename", 1, 0, 'N'},
		{"value", 1, 0, 0},
		{0, 0, 0, 0}
	};

	conf_schedtool_setdefault(conf);

	while (1) {
		int idx, c;
		c = getopt_long(argc, argv, "s:N:h?", long_opts, &idx);
		if (c == -1) {
			if (conf->opmode == CONF_SCHEDTOOL_OPMODE_PRINT) 
			 if (optind < argc)
			  if ((conf->shareval = atoi(argv[optind])) >= CONF_SCHEDTOOL_MINSHAREVAL) conf->opmode=CONF_SCHEDTOOL_OPMODE_ALTER;
			break;
		}

		switch (c) {
		case 0:
			conf->shareval = atoi(optarg);
			if (conf->shareval < CONF_SCHEDTOOL_MINSHAREVAL) {
				fprintf(stderr, "invalid sharevalue\n\n");
				return -1;
			} else conf->opmode=CONF_SCHEDTOOL_OPMODE_ALTER;
			break;
		case 's':
			conf->opmode=CONF_SCHEDTOOL_OPMODE_PRINT;
			break;
		case 'N':
			if (strlen(optarg) < CONF_SCHEDTOOL_SHARENAMEMAXLEN) {
				memset(conf->sharename, 0x0, sizeof(conf->sharename));
				strcpy(conf->sharename, optarg);
			} else {
				fprintf(stderr, "invalid sharename (too long!)\n\n");
				return -1;
			}
			break;
		case '?':
		case 'h':
			usage(stderr);
			return -1;
		default:
			break;
		};

	} // END: WHILE
	

	conf->current_shareval=currentShareValueEx(conf);
	return 0;
}


#ifdef DEBUG_SHARETOOL
static void dumpdefault(FILE* device) {
	fprintf(device, "parsed config DUMP:\n");
	fprintf(device, "   globalconf.sharename          = %s\n", globalconf.sharename);
	fprintf(device, "   globalconf.shareval           = %i\n", globalconf.shareval);

	fprintf(device, "   globalconf.userid             = %i\n", (int)globalconf.userid);
	fprintf(device, "   globalconf.effectiveuserid    = %i\n", (int)globalconf.effectiveuserid);

	fprintf(device, "   globalconf.current_sharevalue = %i\n", globalconf.current_shareval);

	fprintf(device, "   globalconf.opmode             = %i\n", (int)globalconf.opmode);

	fprintf(device, "\n");
}

static void dumpidendity(FILE* device) {
	int  uid = getuid();
	int euid = geteuid();

	fprintf(device, "process idendity DUMP:\n");
	fprintf(device, "   user idendity                 = %i\n", uid);
	fprintf(device, "   user effective idendity       = %i\n", euid);
	fprintf(device, "\n");
}
#endif

int main(int argc, char** argv) {
	if (parseconf(&globalconf, argc, argv) == 0) {
	 /** parsing of command-line parameters SUCCEEDED **/
		fprintf(stderr, SOFTWARE "\n\n");

		#ifdef DEBUG_SHARETOOL
		dumpdefault(stderr);
		dumpidendity(stderr);
		#endif

		if (globalconf.opmode == CONF_SCHEDTOOL_OPMODE_PRINT) {
			fprintf(stderr, "weight of share \"%s\" is: ", globalconf.sharename);
			fflush(stderr);
			if (globalconf.current_shareval < 0)	fprintf(stdout, "0");
			else					fprintf(stdout, "%i",globalconf.current_shareval);
			fflush(stdout);
			fprintf(stderr,"\n\n");
			return globalconf.current_shareval; // return shareval as exitcode !
		} else
		if (globalconf.current_shareval != globalconf.shareval)
		 if (isAllowed()) {
			int result;

			if (globalconf.userid == 0)	result=0;
			else				result=setuid(0);

			if (result == 0) {
				int okay = -1;

				if (updateShareValue() == 0)	okay=0;
				else				fprintf(stderr,"sorry, could not change sharevalue\n");

				// go back from root (NOT EFFECTIVE!)
				if (globalconf.userid != 0) result=setuid(globalconf.userid);
				if (result != 0) fprintf(stderr, "Ohh, can't switch back user...\n");

				if (okay == 0) fprintf(stderr, "new weight of share \"%s\" is: %i\n\n", globalconf.sharename, globalconf.current_shareval);

				return okay;
			} else {
				fprintf(stderr, "can not switch to root!\n\n");
				return -2;
			}
		 } else {
			fprintf(stderr, "you are not allowed to change share \"%s\" to value %i!\n\n", globalconf.sharename, globalconf.shareval);
			return -3;
		 }

	} else {
	 /** parsing of command-line parameters failed **/
		return -4;
	}

	return -5;
}
